// Copyright 2002-2014, University of Colorado Boulder

/**
 * Various methods to aid in the building of the installer package
 *
 * This object is highly dependent upon BitRock installbuilder.
 *
 *
 *
 * @author Michael Kauzmann
 * */

define( function( require ) {
  'use strict';

  // modules
  var winston = require( 'winston' );
  var ShellJS = require( 'shelljs' );
  var fse = require( 'fs-extra' );
  var InstallerBuilderSharedConstants = require( 'InstallerBuilderSharedConstants' );

  // constants
  var INSTALL_BUILDER = InstallerBuilderSharedConstants.INSTALL_BUILDER;


  return {


    /**
     * A generic method that will build the installer for a given platform.
     * @param targetPlatform: either mac, linux, or windows
     * @param coreBuildCommand: the entire command to build the installer from the command line
     */
    buildInstallerForSelectPlatform: function( targetPlatform, coreBuildCommand ) {
      var buildCommand = coreBuildCommand + ' ' + targetPlatform;
      winston.log( 'info', 'Initiating build of installer for ' + targetPlatform + ', command = ' + buildCommand );
      var buildResult = ShellJS.exec( buildCommand );
      if ( buildResult.code !== 0 ) {
        winston.log( 'error', 'Build of installer failed, command result = ' + buildResult.output );
        throw new Error( 'Aborting build due to failure of BitRock installer build step.' );
      }
      winston.log( 'info', 'Build succeeded for ' + targetPlatform + ', buildResult = ' + buildResult.output );

    },


    /**
     * Log the version of the BitRock installer builder.
     * @public
     */
    logInstallerBuilderVersion: function() {

      var versionCommandResult = ShellJS.exec( INSTALL_BUILDER + ' --version' );
      if ( versionCommandResult.code !== 0 ) {
        winston.log( 'error', 'BitRock install builder not found.' );
        throw new Error( 'Aborting build due to inability to find BitRock install builder.' );
      }
      winston.log( 'info', 'BitRock install builder version = ' + versionCommandResult.output );
    },

    /**
     * This method dynamically loads the files and directories for the install builder to package
     *
     * @returns {{directories: Array, files: Array}}
     * directories: A list of the directories to give the install builder
     * files: A list of additional files to give the install builder
     */
    findDistributionsFilesAndDirectories: function() {
      var localWebsiteLocation = InstallerBuilderSharedConstants.RIP_DIR + InstallerBuilderSharedConstants.HOST_NAME + '/';
      var filesInDir = fse.readdirSync( localWebsiteLocation );

      var directories = [];
      var files = [];
      for ( var i in filesInDir ) {
        if ( filesInDir.hasOwnProperty( i ) ) {
          if ( fse.lstatSync( localWebsiteLocation + filesInDir[ i ] ).isDirectory() ) {
            directories.push( filesInDir[ i ] );
          }
          else {
            // There is a weird index[jumple].html like 'index9dfc.html' that shouldn't be added, so select it out.
            if ( (filesInDir[ i ].indexOf( 'index' ) !== -1 && filesInDir[ i ] !== 'index.html' ) ) {
              files.push( filesInDir[ i ] );
            }
          }
        }
      }
      return { directories: directories, files: files };

    }


  };
} );