// Copyright 2002-2015, University of Colorado Boulder

/**
 * Various methods to aid in the ripping of the website
 * Includes the Manipulation of jnlp files and redownloading
 * html5 sims that get screw up with httrack
 * This object is highly dependent upon the ripper utility HTTrack.
 *
 * @author Michael Kauzmann
 */
define( function( require ) {
  'use strict';


  // modules
  var InstallerBuilderSharedConstants = require( 'InstallerBuilderSharedConstants' );
  var JNLPUtils = require( 'JNLPUtils' );
  var ShellJS = require( 'shelljs' );
  var fse = require( 'fs-extra' );
  var http = require( 'http' );
  var winston = require( 'winston' );
  var lookPretty = require( 'eyes' ).inspector();

  // Constants
  var RIPPER = InstallerBuilderSharedConstants.RIPPER;
  var RIP_DIR = InstallerBuilderSharedConstants.RIP_DIR;
  var HOST_NAME = InstallerBuilderSharedConstants.HOST_NAME;
  var WEB_ROOT = InstallerBuilderSharedConstants.WEB_ROOT;

  // HTML
  var HTML_SIMS_LOCAL_LOCATION = RIP_DIR + HOST_NAME + '/sims/html/';
  var HTML_SIMS_LOCAL_NEW_LOCATION_ROOT = InstallerBuilderSharedConstants.TEMP_DIR + 'html5-sims-temp';
  var HTML_SIMS_LOCAL_NEW_LOCATION = HTML_SIMS_LOCAL_NEW_LOCATION_ROOT + '/' + HOST_NAME + '/sims/html/';
  var HTML_SIMS_WEB_LOCATION = WEB_ROOT + 'sims/html/';

  // Java
  var JAVA_SIMS_LOCAL_DIRECTORY = RIP_DIR + HOST_NAME + '/sims/';
  var JAVA_SIMS_WEB_LOCATION = WEB_ROOT + 'sims/';

  // Used to download individual files more efficiently than httrack
  var tryAgain = true;
  var download = function( url, dest, cb ) {
    var file = fse.createWriteStream( dest );
    var request = http.get( url, function( response ) {
      response.pipe( file );
      file.on( 'finish', function() {// Callback once download is complete
        tryAgain = true;
        file.close( cb );  // close() is async, call cb after close completes.

      } );
    } ).on( 'error', function( err ) { // Handle errors
      fse.unlink( dest ); // Delete the file async. (But we don't check the result)
      if ( tryAgain ) {
        // Try downloading again just in case it was a weird network error
        tryAgain = false;
        winston.log( 'error', "Download Failed: " + err.message + ". Trying again." );
        download( url, dest, cb );
      }
      else {
        winston.log( 'error', 'Failed to download: ' + url + "because of: " + err.message );
      }
      if ( cb ) {
        cb();
      }
    } );
  };

  // static object, no constructor
  return {


    /**
     * Conducts the rip command given a configuration
     * @param ripCommand: String of the command to be run in terminal. Including filters and options to guide rip
     * @param cb: callback to be called post rip.
     */
    ripWebsite: function( ripCommand, cb ) {

      winston.log( 'info', 'About to start rip, ripCommand = ' + ripCommand );
      var ripResult = ShellJS.exec( ripCommand );
      if ( ripResult.code !== 0 ) {
        // Rip command failed.  Note the error and abort.
        winston.log( 'error', 'Rip command failed, command output = ' + ripResult.output );
        throw new Error( 'Build aborted due to failure of rip.' );
      }
      winston.log( 'info', 'Initial website rip completed, rip command output = ' + ripResult.output );
      cb();
    },

    /**
     * This will re-download all of the html sims, first puting them in a temp directory,
     * and then copy them over to the original rip location.
     * Then it will make these original folders cleaner by removing older versions of the sims, and
     * random files created when httrack tries to follow links inside the html5 sim.
     *
     * @public
     * */
    fixAllHTML5Sims: function( cb ) {
      var self = this;
      winston.log( 'info', 'Fixing all HTML5 simulations' );

      // I think it's ok to just have this be english
      var userAgent = '"httrack-web-offline-en"';
      var html5RipperOptions = ' -D -%P0 -%e0 -r2 -s0 -A10000000000 --disable-security-limits';

      this.cleanUpHTMLFolders();
      this.removePreviousRip( HTML_SIMS_LOCAL_NEW_LOCATION_ROOT );

      // Rip all correct html sims into a separate folder.
      // This folder is really just to get a dynamic list of all html5 sims on the website,
      // then the 'latest' folder is downloaded for each/
      winston.log( 'info', 'Conducting shallow rip for html sims.' );
      var ripCommand = RIPPER + ' "' + HTML_SIMS_WEB_LOCATION + '" -O "' + HTML_SIMS_LOCAL_NEW_LOCATION_ROOT +
                       '" -F ' + userAgent + ' -%e0 -r2 -s0 -D -%P0 -A10000000000 --disable-security-limits';
      self.ripWebsite( ripCommand, function() {
          var listOfFiles = fse.readdirSync( HTML_SIMS_LOCAL_NEW_LOCATION );
          var listOfSims = [];

          //
          for ( var j in listOfFiles ) {
            if ( listOfFiles.hasOwnProperty( j ) ) {
              if ( fse.lstatSync( HTML_SIMS_LOCAL_NEW_LOCATION + listOfFiles[ j ] ).isDirectory() ) {
                listOfSims.push( listOfFiles[ j ] );
              }
            }
          }

          // Iterate through every html5 sim, rip the latest folder, and copy it into place with the main rip
          for ( var i in listOfSims ) {
            if ( listOfSims.hasOwnProperty( i ) ) {
              self.removePreviousRip( HTML_SIMS_LOCAL_NEW_LOCATION_ROOT );
              winston.log( 'info', 'ripping latest version of html sim: ' + listOfSims[ i ] );
              ripCommand = RIPPER + ' "' + HTML_SIMS_WEB_LOCATION + listOfSims[ i ] + '/latest/' + '" -O "' + HTML_SIMS_LOCAL_NEW_LOCATION_ROOT +
                           '" -F ' + userAgent + html5RipperOptions;
              self.ripWebsite( ripCommand, (function( index ) {
                // Once the sim's latest folder is copied, it is moved to the same path under temp/website
                return function() {
                  winston.log( 'info', 'Copying latest folder of sim: ' + listOfSims[ index ] );
                  // Copy fails unless the directory already exists
                  ShellJS.mkdir( '-p', HTML_SIMS_LOCAL_LOCATION + listOfSims[ index ] + '/latest' );
                  ShellJS.cp( '-Rf', HTML_SIMS_LOCAL_NEW_LOCATION + listOfSims[ index ] + '/latest/*',
                    HTML_SIMS_LOCAL_LOCATION + listOfSims[ index ] + '/latest' );
                };
              }( i ) ) );
            }
          }
          cb();
        }
      );


    },


    /**
     * The html sims get screwed up in the original rip because httrack mangles the sims since it tries to process
     * its links. Therefore the folders contain many useless and unnecessary files
     *
     * Download the HTML5 simulations.  These were already captured during the
     * initial rip, but that rip mangles the sims since it tries to process
     * its links, so they need to be re-ripped with different options.  See
     * Unfuddle issue #3650 for more information.
     */
    cleanUpHTMLFolders: function() {
      winston.log( 'info', "Cleaning up previous rip's html sim folders." );


      var htmlSimFolders = fse.readdirSync( HTML_SIMS_LOCAL_LOCATION );
      // Iterate through each sim folder
      for ( var i in htmlSimFolders ) {
        if ( htmlSimFolders.hasOwnProperty( i ) ) {
          // Check to make sure the file is a directory, mainly running into this problem with 'index.html'
          if ( fse.lstatSync( HTML_SIMS_LOCAL_LOCATION + htmlSimFolders[ i ] ).isDirectory() ) {
            // Read contents of sim folder
            var filesInSimFolder = fse.readdirSync( HTML_SIMS_LOCAL_LOCATION + htmlSimFolders[ i ] );
            //winston.log( 'info', filesInSimFolder );
            for ( var j in filesInSimFolder ) {
              if ( filesInSimFolder.hasOwnProperty( j ) ) {
                // We only want to save the .png files
                if ( filesInSimFolder[ j ].indexOf( 'latest' ) === -1 && filesInSimFolder[ j ].indexOf( 'index' ) === -1 ) {
                  ShellJS.rm( '-rf', HTML_SIMS_LOCAL_LOCATION + htmlSimFolders[ i ] + '/' + filesInSimFolder[ j ] );
                }
              }
            }
            var filesInLatest = fse.readdirSync( HTML_SIMS_LOCAL_LOCATION + htmlSimFolders[ i ] + '/latest' );
            for ( var k in filesInLatest ) {
              if ( filesInLatest.hasOwnProperty( k ) ) {
                // We only want to save the a few files in latest, because it will be replaced eventually
                var possiblyDoomedFile = filesInLatest[ k ];

                // Pretty sure we only need the png, other pages to be uncommented back in if problems are run into
                if ( possiblyDoomedFile.indexOf( '.png' ) === -1 ) { //&& !possiblyDoomedFile.contains('index') &&
                  //       !possiblyDoomedFile.contains('dependencies') && !possiblyDoomedFile.contains('string-map')) {
                  ShellJS.rm( '-rf', HTML_SIMS_LOCAL_LOCATION + htmlSimFolders[ i ] + '/latest/' + possiblyDoomedFile );
                }
              }
            }
          }

        }
      }
    },


    /**
     *  This Function is responsible for manipulating all of the jnlps, and for downloading the installer_all.jars for every java sim
     *  @public
     * */
    fixAllJavaSims: function( cb ) {
      winston.log( 'info', 'Fixing Java Sims' );

      // This method gathers the list of java sims by dynamically searching for any all.jar files
      // If the convention on the website is changed, this method will need to be substantially reworked.
      // Find every all.jar
      var allJarsList = ShellJS.find( JAVA_SIMS_LOCAL_DIRECTORY + '.' ).filter( function( file ) {
        return file.match( /.*all\.jar/ );
      } );//.filter( function( file ) { return !file.match( /\.pack\.gz/ )} );
      // Add all of the jar list to a list of sim names.
      var simNames = [];
      for ( var i = 0; i < allJarsList.length; i++ ) {
        simNames.push( allJarsList[ i ].split( 'phet.colorado.edu/sims/' )[ 1 ].split( '/' )[ 0 ] );
      }
      var numSimsDownloadedAndFixed = 0;
      for ( var j = 0; j < simNames.length; j++ ) {
        this.downloadOneAllJar( simNames[ j ], (function( simName ) {
          return function() {
            var jnlps = ShellJS.find( JAVA_SIMS_LOCAL_DIRECTORY + simName ).filter( function( file ) {
              return file.match( /.*\.jnlp/ );

            } );

            JNLPUtils.handleJNLPs( jnlps );
            numSimsDownloadedAndFixed++;
            // Make sure the call back is called in the correct place.
            if ( numSimsDownloadedAndFixed === simNames.length ) {
              cb();
            }
          };
        }( simNames[ j ] )) );
      }


    },

    /**
     * Helper method for functions that downloads _all_installer.jar files from phet
     * @private
     * */
    downloadOneAllJar: function( simName, cb ) {
      // Will download the _all_installer.jar version of the all jar, then renamed after download

      var INSTALLER_ALL_JAR_NAME = simName + '_all_installer.jar';
      var ALL_JAR_NAME = simName + '_all.jar';
      var localJavaSimLocation = JAVA_SIMS_LOCAL_DIRECTORY + simName + '/';

      download( JAVA_SIMS_WEB_LOCATION + simName + '/' + INSTALLER_ALL_JAR_NAME, localJavaSimLocation + ALL_JAR_NAME,
        function( err ) {
          if ( !err ) {
            winston.log( 'info', 'Copy completed of Java all jar: ' + simName );
            cb();
          }
          else {
            winston.log( 'error', 'Java Sim: ' + INSTALLER_ALL_JAR_NAME + ' copy failed, error: ' + err.message );
            throw new Error( 'Java _all.jar not ripped' );
          }
        } );
    },

    /**
     * Used by the 'quick' version of the ripper to get only one java sim in english
     * @public
     */
    ripOneJavaSimEn: function( simName, cb ) {


      var JNLP_PATH = RIP_DIR + HOST_NAME + '/sims/' + simName + '/';
      var JNLP_FILE = JNLP_PATH + simName + "_en.jnlp";

      this.downloadOneAllJar( simName, function() {
        JNLPUtils.changeCodeBaseAndPermissions( JNLP_FILE, cb );
      } );


    },

    /**
     * Used by the 'quick' version of the ripper to get only one html sim in english
     * @public
     */
    ripOneHTML5SimEn: function( simName, cb ) {

      //    var HTML5_RIP_DIR = RIP_DIR + 'html5Sims/';

      var HTML5Path = simName + '/latest/';
      var HTML5Sim = simName + '_en.html';

      var localHTML5SimLocation = HTML_SIMS_LOCAL_LOCATION + HTML5Path;

      // Check if the sim already exists from a previous rip and remove it.
      winston.log( 'info', 'Removing previous, non-working HTML5sim' );
      ShellJS.rm( '-f', localHTML5SimLocation + HTML5Sim );

      download( WEB_ROOT + HTML5Path + HTML5Sim, localHTML5SimLocation + HTML5Sim,
        function( err ) {
          if ( !err ) {
            winston.log( 'info', 'Copy completed of HTML5 sim: ' + simName );
            cb();
          }
          else {
            winston.log( 'error', 'HTML5 Sim: ' + simName + ' copy failed, error: ' + err.message );
            throw new Error( 'HTML5 sim not ripped' );
          }
        } );
    },

    /**
     * The rip generates many random looking files that serve no purpose
     * and clutter up the project, so they should be removed.
     * This method keeps all 'index.html', 'register.html' , 'sign-in.html'
     * files while removing the strange and seemingly useless weird numbered versions of these files
     *
     *  TODO we may want to consider doing this in httrack filters, rather than post processing.
     */
    removeExtraHTMLFiles: function() {
      var extraFileTypes = [ 'index', 'register', 'sign-in', 'search' ];
      for ( var j in extraFileTypes ) {
        if ( extraFileTypes.hasOwnProperty( j ) ) {
          var extraIndexFiles = (function( index ) {
            return ShellJS.find( RIP_DIR + HOST_NAME + '/.' )
              .filter( function( file ) {return file.match( new RegExp( '.*' + extraFileTypes[ index ] + '.*\\.html' ) );} )
              .filter( function( file ) { return !file.match( new RegExp( extraFileTypes[ index ] + '\\.html' ) );} );
          }( j ));
          for ( var i in extraIndexFiles ) {
            if ( extraIndexFiles.hasOwnProperty( i ) ) {
              ShellJS.rm( '-f', extraIndexFiles[ i ] );
            }
          }
        }
      }
    },


    /**
     * See if a previous rip of the web site is present and, if so, remove it.
     * This is so that httrack does not get confused and think we are updating the previous rip.
     * @param folder: optional, path of folder to be deleted, default is InstallerBuilderSharedConstants.RIP_DIR
     * @private
     */


    removePreviousRip: function( folder ) {
      if ( !!folder ) {
        winston.log( 'info', 'Folder to be removed: ' + folder );
      }
      folder = folder || InstallerBuilderSharedConstants.RIP_DIR;
      winston.log( 'info', 'Checking if previous rip exists...' );
      if ( fse.existsSync( folder ) ) {
        winston.log( 'info', 'Previous rip found, removing it...' );
        fse.removeSync( folder );
        winston.log( 'info', 'Previous rip successfully removed.' );
      }
      else {
        winston.log( 'info', 'No previous rip detected, moving on.' );
      }
    },


    /**
     * Appends all locales wanted in this rip to the user agent.
     * This is done so that httrack knows which languages to follow links to.
     * @param withActivities - boolean
     * @param locales - list of the supported locales
     * @returns {string}
     */
    setUserAgent: function( withActivities, locales ) {
      var userAgent = '';
      if ( withActivities ) {
        userAgent = ' -F "httrack-web-offline-activities-en';
      }
      else {
        userAgent = ' -F "httrack-web-offline-en';
      }
      for ( var i in locales ) {
        if ( locales.hasOwnProperty( i ) ) {
          if ( locales[ i ] !== 'en' ) {
            userAgent += ',' + locales[ i ];
          }
        }
      }
      userAgent += '" ';
      return userAgent;
    },


    /**
     * used by temp-harness.js for small and simple tests
     *
     */
    tempTest: function() {
      this.removeExtraIndexFiles();

    }

  }
    ;
} )
;